from typing import Iterable, List, Any

__all__ = ['Language']


class Language(object):
    """
    Represents a programming language that is supported by boggart.
    """
    @staticmethod
    def from_dict(d: dict) -> 'Language':
        assert 'name' in d
        assert 'file-endings' in d
        assert isinstance(d['name'], str)
        assert isinstance(d['file-endings'], list)
        assert all(isinstance(e, str) for e in d['file-endings'])

        name = d['name']
        file_endings = d['file-endings']
        return Language(name, file_endings)

    def __init__(self,
                 name: str,
                 file_endings: List[str]
                 ) -> None:
        self.__name = name
        self.__file_endings = frozenset(file_endings)

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Language) and \
               self.name == other.name and \
               list(self.file_endings) == list(other.file_endings)

    def __repr__(self) -> str:
        return "Language({self.name})".format(self=self)

    @property
    def name(self) -> str:
        """
        The name of the language.
        """
        return self.__name

    @property
    def file_endings(self) -> Iterable[str]:
        """
        A list of known file endings used by this language. These are used to
        automatically detect the language used by a given file when language
        information is not explicitly provided.
        """
        return self.__file_endings.__iter__()

    def to_dict(self) -> dict:
        """
        Provides a dictionary-based description of this language, ready to be
        serialized.
        """
        return {
            'name': self.name,
            'file-endings': list(self.file_endings)
        }



from typing import Iterator, Any, List, FrozenSet, Optional, Dict, Set
import warnings
import os
import logging

from ..core import Language
from ..exceptions import IllegalConfig, LanguageNotDetected
from ..warnings import LanguageOverwriteWarning

logger = logging.getLogger(__name__)

__all__ = ['Languages']


class Languages(object):
    """
    Maintains information about the languages that are supported by Hulk.
    """
    @staticmethod
    def from_defs(defs: List[Any], base: 'Languages' = None) -> 'Languages':
        """
        Loads an operator configuration from a list of definitions taken
        from a configuration file, together with an optionally provided
        parent (overall) configuration.
        """
        logger.info("Loading languages from definitions: %s", defs)
        config = base if base else Languages()
        for d in defs:
            logger.info("Loading language from definition: %s", d)
            language = Language.from_dict(d)
            logger.info("Adding loaded language to configuration: %s",
                        language)
            config = config.add(language)
        logger.info("Loaded languages from definitions.")
        return config

    def __init__(self,
                 languages: Optional[Dict[str, Language]] = None
                 ) -> None:
        """
        Constructs a collection of lamnguages for a set of languages provided
        in the form a dictionary, indexed by name.
        """
        self.__languages = dict(languages) if languages else {}

    def add(self, language: Language) -> 'Languages':
        """
        Returns a variant of this collection of languages that also includes a
        given language.
        """
        logger.info("Added language to collection: %s", language)
        endings = set(self.supported_file_endings)

        # if there already exists a language with the given name, produce a
        # warning and remove its file endings from consideration.
        if language.name in self.__languages:
            old_version = self.__languages[language.name]
            msg = "config overwrites existing language definition: {}."
            msg = msg.format(language.name)
            warnings.warn(msg, LanguageOverwriteWarning)
            endings -= set(old_version.file_endings)

        # are the file endings used by this language already in use?
        if set(language.file_endings) & endings:
            logger.error("failed to load language due to an existing language sharing a common file ending.")  # noqa: pycodestyle
            raise IllegalConfig("file ending ambiguity: two or more languages share a common file ending.")  # noqa: pycodestyle

        languages = dict(self.__languages)
        languages[language.name] = language
        logger.info("Added language to collection: %s", language)
        return Languages(languages)

    def __iter__(self) -> Iterator[Language]:
        """
        An iterator over the languages within this collection.
        """
        for name in self.__languages:
            yield self.__languages[name]

    def __getitem__(self, name: str) -> Language:
        """
        Attempts to fetch the definition of the language associated with a
        supplied name.

        Raises:
            KeyError: if no language is found with the given name.
        """
        return self.__languages[name]

    def __contains__(self, name: str) -> bool:
        """
        Provides an alias for `supports`.
        """
        return self.supports(name)

    def __len__(self) -> int:
        """
        Returns the number of languages contained within this collection.
        """
        return len(self.__languages)

    def supports(self, name: str) -> bool:
        """
        Determines whether this colelctions contains a language with a given
        name.
        """
        return name in self.__languages

    @property
    def supported_file_endings(self) -> FrozenSet[str]:
        """
        The set of file endings that are used by languages within this
        collection.
        """
        endings = set()  # type: Set[str]
        for language in self:
            endings.union(language.file_endings)
        return frozenset(endings)

    def detect(self, filename: str) -> Language:
        """
        Attempts to automatically detect the language used by a file based
        on its file ending.

        Returns:
            The language associated with the file ending used by the given
            file.

        Raises:
            LanguageNotDetected: if the language used by the filename could not
                be automatically detected.
        """
        logger.info("Attempting to detect language used by file: %s",
                    filename)
        _, suffix = os.path.splitext(filename)
        logger.debug("Using suffix of file, '%s': '%s'",
                     filename, suffix)
        for language in self:
            if suffix in language.file_endings:
                logger.info("Detected language used by file, '%s': %s",
                            filename, language)
                return language
        logger.error("Failed to detect language used by file: %s", filename)
        raise LanguageNotDetected(filename)



from typing import Iterator, Dict

from .api import API
from ..core import Operator


class OperatorCollection(object):
    """
    Provides read-only access to the set of operator supported by a given
    server.
    """
    def __init__(self, api: API) -> None:
        """
        Constructs a new language collection for a given server.

        Parameters:
            api: the low-level client API that should be used to interact
                with the server.
        """
        dict_list = api.get("/operators").json()
        operators = [Operator.from_dict(d) for d in dict_list]
        self.__contents = \
            {op.name: op for op in operators}  # type: Dict[str, Operator]

    def __iter__(self) -> Iterator[Operator]:
        """
        Returns an iterator over the operators within this collection.
        """
        return self.__contents.values().__iter__()

    def __len__(self) -> int:
        """
        Returns a count of the number of operators within this collection.
        """
        return len(self.__contents)

    def __getitem__(self, name: str) -> Operator:
        """
        Returns the operator associated with a given name.

        Raises:
            KeyError: if no operator is found with the given name.
        """
        return self.__contents[name]



from typing import Optional
import logging

import yaml

from .languages import Languages
from .operators import Operators
from ..exceptions import BadConfigFile

logger = logging.getLogger(__name__)

__all__ = ['Configuration']


class Configuration(object):
    @staticmethod
    def from_file(filename: str,
                  parent: Optional['Configuration'] = None
                  ) -> 'Configuration':
        """
        Loads a configuration file from a given file.
        """
        logger.info("Loading configuration from file: %s", filename)
        config = parent if parent else Configuration()
        logger.info("Using parent configuration: %s", config)
        languages = config.languages
        operators = config.operators

        logger.debug("Attempting to read contents of config file: %s",
                     filename)
        with open(filename, 'r') as f:
            yml = yaml.load(f)
        logger.debug("Read YAML contents of config file: %s",
                     filename)

        if 'version' not in yml:
            logger.error("Bad configuration file: missing 'version' property.")
            raise BadConfigFile("expected 'version' property")
        if yml['version'] != '1.0':
            logger.error("Bad configuration file: unsupported version.")
            raise BadConfigFile("unexpected 'version' property; only '1.0' is currently supported.")  # noqa: pycodestyle

        # update the languages and operators provided by this configuration
        logger.debug("Loading languages from config file.")
        languages = \
            Languages.from_defs(yml.get('languages', []), languages)
        logger.debug("Loaded languages from config file.")
        logger.debug("Loading operators from config file.")
        operators = \
            Operators.from_defs(yml.get('operators', []),
                                languages=languages,
                                base=operators)
        logger.debug("Loaded operators from config file.")
        logger.info("Loaded configuration from file: %s", filename)
        return Configuration(languages, operators)

    def __init__(self,
                 languages: Optional[Languages] = None,
                 operators: Optional[Operators] = None
                 ) -> None:
        """
        Constructs a configuration that supports a given set of operators and
        languages.
        """
        self.__languages = languages if languages else Languages()
        self.__operators = operators if operators else Operators()

    @property
    def languages(self) -> Languages:
        """
        The languages defined by this configuration.
        """
        return self.__languages

    @property
    def operators(self) -> Operators:
        """
        The mutation operators defined by this configuration.
        """
        return self.__operators



# TODO mutants are killed when the server is killed
from typing import List, Iterator, Dict
from uuid import UUID, uuid4
import tempfile
import logging

from bugzoo.core.bug import Bug
from bugzoo.core.patch import Patch
from bugzoo.client import Client as BugZooClient
from rooibos import Client as RooibosClient

from .sourcefile import SourceFileManager
from ..config.operators import Operators as OperatorManager
from ..core import Mutant, Mutation, Replacement

logger = logging.getLogger(__name__)


class MutantManager(object):
    def __init__(self,
                 client_bugzoo: BugZooClient,
                 client_rooibos: RooibosClient,
                 operators: OperatorManager,
                 sources: SourceFileManager
                 ) -> None:
        self.__mutants = {}  # type: Dict[UUID, Mutant]
        self.__bugzoo = client_bugzoo
        self.__rooibos = client_rooibos
        self.__operators = operators
        self.__sources = sources

    def __iter__(self) -> Iterator[UUID]:
        """
        Returns an iterator over the UUIDs of the mutants that are currently
        registered with this server.
        """
        yield from self.__mutants.keys()

    def __getitem__(self, uuid: UUID) -> Mutant:
        """
        Retrieves a registered mutant by its UUID.

        Raises:
            KeyError: if no mutant is registered under the given UUID.
        """
        return self.__mutants[uuid]

    def __delitem__(self, uuid: UUID) -> None:
        """
        Deletes a registered mutant and destroys its allocated resources.

        Raises:
            KeyError: if no mutant is registered under the given UUID.
        """
        logger.info("Attempting to destroy mutant with UUID: %s", uuid.hex)
        try:
            mutant = self[uuid]
        except KeyError:
            logger.exception("Failed to find mutant with UUID: %s", uuid.hex)
            raise
        try:
            self.__bugzoo.docker.delete_image(mutant.docker_image)
        except Exception:
            logger.exception("Failed to destroy docker image (%s) for mutant: %s",  # noqa: pycodestyle
                             mutant.docker_image,
                             mutant)
        # FIXME deregister the BugZoo snapshot
        del self.__mutants[uuid]

    def __len__(self) -> int:
        """
        Returns a count of the number of registered mutants.
            """
        return len(self.__mutants)

    def generate(self, snapshot: Bug, mutations: List[Mutation]) -> Mutant:
        """
        Generates a mutant by applying a given set of mutations to a BugZoo
        snapshot.

        Parameters:
            snapshot: the BugZoo snapshot to mutate.
            mutations: the mutations to apply to the snapshot.

        Returns:
            a description of the generated mutant.
        """
        logger.info("Generating mutant of snapshot '%s' by applying mutations: %s",  # noqa: pycodestyle
                    snapshot.name,
                    ', '.join([repr(m) for m in mutations]))
        bz = self.__bugzoo
        assert len(mutations) <= 1, \
            "higher-order mutation is currently unsupported"

        # NOTE this is *incredibly* unlikely to conflict
        logging.debug("generating UUID for mutant...")
        uuid = uuid4()
        logger.debug("generated UUID for mutant: %s", uuid.hex)
        try:
            assert uuid not in self.__mutants, "UUID already in use."
        except AssertionError:
            logger.exception("automatically generated UUID is already in use: %s",  # noqa: pycodestyle
                             uuid)
            raise
        logger.debug("constructing mutation description...")
        mutant = Mutant(uuid, snapshot.name, mutations)
        logger.debug("constructed mutant description: %s", mutant)

        # generate a diff for the mutant
        logger.debug("generating a unified diff for mutant")
        diff = \
            self.__sources.mutations_to_diff(snapshot, list(mutant.mutations))
        logger.debug("generated unified diff for mutant")

        # generate the Docker image on the BugZoo server
        logger.debug("provisioning container to persist mutant as a snapshot")
        container = bz.containers.provision(snapshot)
        logger.debug("provisioned container [%s] for mutant [%s].",
                     container.uid, mutant.uuid.hex)
        try:
            logger.debug("applying mutation patch to original source code.")
            bz.containers.patch(container, diff)
            logger.debug("applied mutation patch to original source code.")
            bz.containers.persist(container, mutant.docker_image)
        finally:
            del bz.containers[container.uid]
            logger.debug("destroyed temporary container [%s] for mutant [%s].",
                         container.uid, mutant.uuid.hex)

        # build and register a BugZoo snapshot
        files_to_instrument = snapshot.files_to_instrument
        snapshot_mutated = Bug(name=mutant.snapshot,
                               image=mutant.docker_image,
                               program=snapshot.program,
                               dataset=None,
                               source=None,
                               source_dir=snapshot.source_dir,
                               languages=snapshot.languages,
                               harness=snapshot.harness,
                               compiler=snapshot.compiler,
                               files_to_instrument=files_to_instrument)
        logger.debug("Registering snapshot for mutant with BugZoo: %s",
                     mutant.uuid.hex)
        bz.bugs.register(snapshot_mutated)
        logger.debug("Registered snapshot for mutant with BugZoo: %s",
                     mutant.uuid.hex)

        # track the mutant
        logger.debug("Registering mutant with UUID '%s': %s",
                     mutant.uuid.hex,
                     mutant)
        self.__mutants[mutant.uuid] = mutant
        logger.debug("Registered mutant with UUID '%s'", mutant.uuid.hex)
        logger.info("Generated mutant: %s", mutant)
        return mutant



from typing import Optional, Tuple, Dict, Iterator, List
import os
import logging

import rooibos
from bugzoo.client import Client as BugZooClient
from bugzoo.core.bug import Bug
from bugzoo.core.fileline import FileLine
from rooibos import Client as RooibosClient

from .mutant import MutantManager
from .sourcefile import SourceFileManager
from ..exceptions import *
from ..core import Language, Mutation, Operator, Mutant, FileLocationRange, \
                   Location
from ..config import Configuration, Languages, Operators

logger = logging.getLogger(__name__)

__all__ = ['Installation']


class Installation(object):
    """
    Used to manage a local installation of boggart.
    """
    @classmethod
    def default_user_config_path(cls) -> str:
        """
        The default path to the user-level configuration file.
        """
        if 'BOGGART_USER_CONFIG_PATH' in os.environ:
            return os.environ['BOGGART_USER_CONFIG_PATH']

        home = os.environ['HOME']
        default_path = os.path.join(home, '.boggart.yml')
        return default_path

    @classmethod
    def sys_config_path(cls) -> str:
        """
        The path to the system-level configuration file.
        """
        src_dir = os.path.dirname(__file__)
        cfg_fn = os.path.join(src_dir, '../config/sys.boggart.yml')
        return cfg_fn

    @classmethod
    def load(cls,
             client_bugzoo: BugZooClient,
             client_rooibos: RooibosClient,
             *,
             user_config_path: Optional[str] = None
             ) -> 'Installation':
        """
        Loads a boggart installation.

        Parameters:
            client_bugzoo: A connection to the BugZoo server that should be
                used by this boggart server.
            config_filepath: The path to the user configuration file for
                boggart. If left unspecified,
                `Installation.default_user_config_path` will be used instead.
        """
        logger.info("loading boggart installation")
        if not user_config_path:
            user_config_path = Installation.default_user_config_path()
            logger.info("no user config path provided -- using default user config path, '%s', instead",  # noqa: pycodestyle
                        user_config_path)

        logger.info("loading system configuration from file: %s",
                    cls.sys_config_path())
        system_cfg = Configuration.from_file(Installation.sys_config_path())
        logger.info("loading system configuration")

        if not os.path.isfile(user_config_path):
            logger.info("no user configuration file found at %s",
                        user_config_path)
            return Installation(system_cfg, client_bugzoo, client_rooibos)

        logger.info("loading user configuration from file: %s",
                    user_config_path)
        user_cfg = Configuration.from_file(user_config_path, system_cfg)
        logger.info("loaded user configuration from file: %s",
                    user_config_path)
        return Installation(user_cfg, client_bugzoo, client_rooibos)

    def __init__(self,
                 config: Configuration,
                 client_bugzoo: BugZooClient,
                 client_rooibos: RooibosClient
                 ) -> None:
        self.__config = config
        self.__bugzoo = client_bugzoo
        self.__rooibos = client_rooibos
        self.__sources = SourceFileManager(client_bugzoo,
                                           client_rooibos,
                                           config.operators)
        self.__mutants = MutantManager(client_bugzoo,
                                       client_rooibos,
                                       config.operators,
                                       self.__sources)

    @property
    def bugzoo(self) -> BugZooClient:
        """
        A connection to the BugZoo server to which this boggart server is
        attached.
        """
        return self.__bugzoo

    @property
    def rooibos(self) -> RooibosClient:
        """
        A connection to the Rooibos server to which this boggart server is
        attached.
        """
        return self.__rooibos

    @property
    def sources(self) -> SourceFileManager:
        """
        Provides access to the contents and caches the contents of source code
        belonging to BugZoo snapshots.
        """
        return self.__sources

    @property
    def languages(self) -> Languages:
        """
        The languages registered with this installation.
        """
        return self.__config.languages

    @property
    def operators(self) -> Operators:
        """
        The mutation operators registered with this installation.
        """
        return self.__config.operators

    @property
    def mutants(self) -> MutantManager:
        """
        The active mutants that are registered with this installation.
        """
        return self.__mutants

    def mutations(self,
                  snapshot: Bug,
                  filepath: str,
                  *,
                  language: Language = None,
                  operators: List[Operator] = None,
                  restrict_to_lines: Iterator[FileLine] = None
                  ) -> Iterator[Mutation]:
        """
        Computes all of the first-order mutations that can be applied to a
        given file belonging to a specified BugZoo snapshot.

        Returns:
            an iterator over the possible mutations.

        Raises:
            LanguageNotDetected: if no language is specified and the language
                used by the file cannot be automatically determined.
            FileNotFound: if the given file is not found inside the snapshot.
        """
        logger.info("Computing mutations to file, '%s', in snapshot, '%s'",
                    filepath,
                    snapshot.name)

        def match_to_mutation(op_name: str,
                              transformation_index: int,
                              match: rooibos.Match
                              ) -> Mutation:
            logger.debug("Transforming template match to mutation: %s", match)
            start = Location(match.location.start.line,
                             match.location.start.col)
            stop = Location(match.location.stop.line,
                            match.location.stop.col)
            location = FileLocationRange(filepath, start, stop)

            args = {}  # type: Dict[str, str]
            for term in match.environment:
                value = match.environment[term].fragment
                args[term] = value

            mut = Mutation(op_name, transformation_index, location, args)
            logger.debug("Transformed match, %s, to mutation: %s", match, mut)
            return mut

        if operators is None:
            logger.info("No mutation operators specified -- attempting to use all available operators.")  # noqa: pycodestyle
            operators = list(self.operators)
        logger.info("Using mutation operators: %s",
                    ', '.join([op.name for op in operators]))

        logger.info("Obtaining source code for specified file: %s", filepath)
        text = self.sources.read_file(snapshot, filepath)
        logger.info("Obtained source code for file %s:\n%s", filepath, text)

        if language is None:
            logger.info("Attempting to automatically detect language used by file: %s",  # noqa: pycodestyle
                        filepath)
            language = self.languages.detect(filepath)
        logger.info("Treating '%s' as a %s file.", filepath, language.name)

        for operator in operators:
            logger.info("Using operator to find mutations: %s", operator.name)
            for (idx, transformation) in enumerate(operator.transformations):
                logger.debug("Finding all instances of match template in source code: %s",  # noqa: pycodestyle
                             transformation.match)
                for match in self.rooibos.matches(text, transformation.match):
                    logger.debug("Found template match: %s", match)
                    yield match_to_mutation(operator.name, idx, match)



from typing import Dict, Tuple, List
from difflib import unified_diff
import logging

from bugzoo.core.patch import Patch
from bugzoo.core.bug import Bug
from bugzoo.client import Client as BugZooClient
from rooibos import Client as RooibosClient

from ..config.operators import Operators as OperatorManager
from ..core import FileLocationRange, Replacement, Mutation, FileLine, Location
from ..exceptions import *

logger = logging.getLogger(__name__)

__all__ = ['SourceFileManager']


class SourceFileManager(object):
    def __init__(self,
                 client_bugzoo: BugZooClient,
                 client_rooibos: RooibosClient,
                 operators: OperatorManager
                 ) -> None:
        self.__bugzoo = client_bugzoo
        self.__rooibos = client_rooibos
        self.__operators = operators
        self.__cache_file_contents = {}  # type: Dict[Tuple[str, str], str]
        self.__cache_offsets = {}  # type: Dict[Tuple[str, str], List[int]]

    def num_lines(self, snapshot: Bug, filepath: str) -> int:
        """
        Computes the number of lines that belong to a particular file in a
        given snapshot.

        Raises:
            FileNotFound: if the given file is not found within the provided
                snapshot.
        """
        return len(self._line_offsets(snapshot, filepath))

    def _forget_file(self, snapshot: Bug, filepath: str) -> None:
        """
        Removes all stored information for a particular file belonging to a
        given snapshot.
        """
        try:
            cache_key = (snapshot.name, filepath)
            del self.__cache_offsets[cache_key]
            del self.__cache_file_contents[cache_key]
        except KeyError:
            pass

    def _line_offsets(self, snapshot: Bug, filepath: str) -> List[int]:
        """
        Returns a list specifying the offset for the first character on each
        line in a given file belonging to a BugZoo snapshot.
        """
        logger.debug("Fetching line offsets for file, '%s', in snapshot, '%s'",  # noqa: pycodestyle
                     filepath,
                     snapshot.name)
        key_cache = (snapshot.name, filepath)
        if key_cache in self.__cache_offsets:
            logger.debug("Retrieving line offsets for file, '%s', in snapshot, '%s', from cache.",  # noqa: pycodestyle
                         filepath,
                         snapshot.name)
            return self.__cache_offsets[key_cache]

        logger.debug("Computing line offsets for file, '%s', in snapshot, '%s'",  # noqa: pycodestyle
                     filepath,
                     snapshot.name)
        contents = self.read_file(snapshot, filepath)

        # find all indices of newline characters
        offsets = [0]
        last_offset = 0
        while True:
            next_line_break = contents.find('\n', last_offset)
            if next_line_break == -1:
                break
            last_offset = next_line_break + 1
            offsets.append(last_offset)

        logger.debug("Saving line offsets for file, '%s', in snapshot, '%s', to cache.",  # noqa: pycodestyle
                     filepath,
                     snapshot.name)
        self.__cache_offsets[key_cache] = offsets
        return offsets

    def line_col_to_offset(self,
                           snapshot: Bug,
                           filepath: str,
                           line_num: int,
                           col_num: int
                           ) -> int:
        """
        Transforms a line-column number for a given file belonging to a
        BugZoo snapshot into a zero-indexed character offset.
        """
        assert line_num > 0
        assert col_num >= 0
        line_col_s = "%s/%s[%d:%d]".format(snapshot.name,
                                           filepath,
                                           line_num,
                                           col_num)
        logger.debug("Transforming line-column, '%s', into a character offset",  # noqa: pycodestyle
                     line_col_s)
        line_offsets = self._line_offsets(snapshot, filepath)
        line_starts_at = line_offsets[line_num - 1]
        offset = line_starts_at + col_num
        logger.debug("Transformed line-column, '%s', into character offset: %s",  # noqa: pycodestyle
                     line_col_s,
                     offset)
        return offset

    def mutations_to_diff(self,
                          snapshot: Bug,
                          mutations: List[Mutation]
                          ) -> Patch:
        """
        Transforms a list of mutations to a given snapshot into a unified diff
        that can be applied to that snapshot.
        """
        replacements = self.mutations_to_replacements(snapshot, mutations)
        diff = self.replacements_to_diff(snapshot, replacements)
        return diff

    def mutations_to_replacements(self,
                                  snapshot: Bug,
                                  mutations: List[Mutation]
                                  ) -> Dict[str, List[Replacement]]:
        """
        Transforms a list of mutations to a given snapshot into a set of
        source code replacements.

        Parameters:
            snapshot: the snapshot to which the mutations should be applied.
            mutations: the mutations to apply to the snapshot.

        Returns:
            a mapping from (modified) files in the snapshot, given by their
            paths relative to the source directory for the snapshot, to a list
            of source code replacements in that file.
        """
        logger.debug("transforming mutations into replacements")
        file_to_replacements = {}  # type: Dict[str, List[Replacement]]
        for mutation in mutations:
            fn = mutation.location.filename
            if fn not in file_to_replacements:
                file_to_replacements[fn] = []
            replacement = self.mutation_to_replacement(snapshot, mutation)
            file_to_replacements[fn].append(replacement)
        logger.debug("transformed mutations to replacements: %s",
                     file_to_replacements)
        return file_to_replacements

    def mutation_to_replacement(self,
                                snapshot: Bug,
                                mutation: Mutation
                                ) -> Replacement:
        """
        Transforms a given mutation to a snapshot into a replacement.
        """
        logger.debug("transforming mutation [%s] to replacement.", mutation)
        operator = self.__operators[mutation.operator]
        transformation = \
            operator.transformations[mutation.transformation_index]
        text_mutated = self.__rooibos.substitute(transformation.rewrite,
                                                 mutation.arguments)
        replacement = Replacement(mutation.location, text_mutated)
        logger.debug("transformed mutation [%s] to replacement [%s].",
                     mutation, replacement)
        return replacement

    def replacements_to_diff(self,
                             snapshot: Bug,
                             file_to_replacements: Dict[str, List[Replacement]]
                             ) -> Patch:
        """
        Transforms a set of replacements into a unified diff for a given
        snapshot.

        Parameters:
            snapshot: the snapshot to which the replacements should be
                applied.
            file_to_replacements: a mapping from files, indexed by their path
                relative to the source directory for the snapshot, to
                replacements in that file.

        Returns:
            a unified diff that applies all of the given replacements to the
            source code for the given snapshot.
        """
        logger.debug("transforming replacements to diff")
        file_diffs = []  # type: List[str]
        for (filename, replacements) in file_to_replacements.items():
            original = self.read_file(snapshot, filename)
            mutated = self.apply(snapshot, filename, replacements)
            diff = ''.join(unified_diff(original.splitlines(True),
                                        mutated.splitlines(True),
                                        filename, filename))
            logger.debug("transformed replacements to file to diff:\n%s", diff)
            file_diffs.append(diff)
        diff_s = '\n'.join(file_diffs)
        logger.debug("transformed mutations to diff:\n%s", diff_s)
        diff = Patch.from_unidiff('\n'.join(file_diffs))
        return diff

    def read_file(self, snapshot: Bug, filepath: str) -> str:
        """
        Fetches the contents of a specified source code file belonging to a
        given BugZoo snapshot.

        Raises:
            FileNotFound: if the given file is not found inside the snapshot.
        """
        logger.debug("Reading contents of source file: %s/%s",  # noqa: pycodestyle
                     snapshot.name, filepath)
        # TODO normalise file path

        bgz = self.__bugzoo
        key_cache = (snapshot.name, filepath)
        if key_cache in self.__cache_file_contents:
            contents = self.__cache_file_contents[key_cache]
            logger.debug("Found contents of source file, '%s/%s', in cache: %s",  # noqa: pycodestyle
                         snapshot.name, filepath, contents)
            return contents

        logger.debug("Provisioning a temporary container to fetch contents of file")  # noqa: pycodestyle
        container = bgz.containers.provision(snapshot)
        try:
            contents = bgz.files.read(container, filepath)
        except KeyError:
            logger.exception("Failed to read source file, '%s/%s': file not found",  # noqa: pycodestyle
                             snapshot.name, filepath)
            raise FileNotFound(filepath)
        finally:
            del bgz.containers[container.uid]
        logger.debug("Read contents of source file, '%s/%s':\n%s",  # noqa: pycodestyle
                     snapshot.name, filepath, contents)

        self.__cache_file_contents[key_cache] = contents
        return contents

    def read_line(self,
                  snapshot: Bug,
                  location: FileLine,
                  *,
                  keep_newline: bool = False
                  ) -> str:
        """
        Fetches the contents of a given line belonging to a particular
        BugZoo snapshot.

        Raises:
            FileNotFound: if the file associated with the given line is
                not found inside the snapshot.

        TODO: LineNotFound
        """
        content_file = self.read_file(snapshot, location.filename)
        line_offsets = self._line_offsets(snapshot, location.filename)
        start_at = line_offsets[location.num - 1]
        if len(line_offsets) == location.num:
            content_line = content_file[start_at:]
        else:
            end_at = line_offsets[location.num]
            content_line = content_file[start_at:end_at]
        return content_line if keep_newline else content_line.rstrip('\n')

    def read_chars(self, snapshot: Bug, location: FileLocationRange) -> str:
        """
        Fetches a specified sequence of characters from a source code file
        belonging to a BugZoo snapshot.

        Raises:
            FileNotFound: if the given file is not found inside the snapshot.
        """
        logger.debug("Reading characters at %s in snapshot, %s",
                     location, snapshot.name)
        filename = location.filename
        contents_file = self.read_file(snapshot, filename)

        start_at = self.line_col_to_offset(snapshot,
                                           filename,
                                           location.start.line,
                                           location.start.column)
        stop_at = self.line_col_to_offset(snapshot,
                                          filename,
                                          location.stop.line,
                                          location.stop.column)

        contents = contents_file[start_at:stop_at + 1]
        logger.debug("Read characters at %s in snapshot, %s: %s",
                     location, snapshot.name, contents)
        return contents

    def apply(self,
              snapshot: Bug,
              filename: str,
              replacements: List[Replacement]
              ) -> str:
        # TODO ensure all replacements are in the same file
        # TODO sort replacements by the start of their affected character range
        # TODO ensure no replacements are conflicting
        logger.debug("applying replacements to source file, '%s/%s': %s",
                     snapshot.name, filename, replacements)
        content = self.read_file(snapshot, filename)
        for replacement in replacements:
            location = replacement.location
            start_at = self.line_col_to_offset(snapshot,
                                               filename,
                                               location.start.line,
                                               location.start.column)
            stop_at = self.line_col_to_offset(snapshot,
                                              filename,
                                              location.stop.line,
                                              location.stop.column)
            content = \
                content[:start_at] + replacement.text + content[stop_at:]
        logger.debug("applied replacements to source file, '%s/%s': %s:\n%s",
                     snapshot.name, filename,
                     replacements,
                     content)
        return content



#
# TODO use metaprogramming to reduce the amount of boilerplate
#
from typing import Any, Dict, Tuple

import flask

__all__ = [
    'BoggartException',
    'UnexpectedResponse',
    'ClientServerError',
    'OperatorNameAlreadyExists',
    'OperatorNotFound',
    'LanguageNotFound',
    'LanguageNotDetected',
    'BadConfigFile',
    'IllegalConfig',
    'FileNotFound',
    'SnapshotNotFound',
    'ConnectionFailure',
    'BadFormat',
    'UnexpectedServerError'
]


class BoggartException(Exception):
    """
    Base class for all exceptions produced by Boggart.
    """
    def __init__(self, message: str) -> None:
        self.__message = message
        super().__init__(message)

    @property
    def message(self):
        """
        A short description of the error.
        """
        return self.__message


class ConnectionFailure(BoggartException):
    """
    Failed to establish a connection to the server within the timeout.
    """
    def __init__(self) -> None:
        super().__init__(
            "failed to establish connection to server within timeout."
        )


class UnexpectedResponse(BoggartException):
    """
    The server produced an unexpected response that the client does not know
    how to interpret or decode.
    """
    def __init__(self, response: flask.Response) -> None:
        msg = "Server produced an unexpected response [{}]:\n{}"
        msg = msg.format(response.status_code, response.text)
        super().__init__(msg)


class ClientServerError(BoggartException):
    """
    Base class for all exceptions that may be thrown from the server to the
    client. Provides the necessary machinery to write/read errors to/from
    JSON descriptions.
    """
    @staticmethod
    def from_dict(d: dict) -> 'BoggartException':
        assert 'error' in d
        d = d['error']

        assert isinstance(d, dict)
        assert 'kind' in d

        # TODO use metaprogramming to avoid having to maintain this list
        cls = ({
            'OperatorNameAlreadyExists': OperatorNameAlreadyExists,
            'LanguageNotFound': LanguageNotFound,
            'BadConfigFile': BadConfigFile,
            'IllegalConfig': IllegalConfig,
            'UnexpectedServerError': UnexpectedServerError
        })[d['kind']]

        return cls.from_data(d.get('data', {}))  # type: ignore

    def __init__(self, status_code: int, message: str) -> None:
        self.__status_code = status_code
        super().__init__(message)

    @property
    def status_code(self) -> int:
        """
        The status code that was produced by the server when this error was
        reported to the client.
        """
        return self.__status_code

    @property
    def data(self) -> Dict[str, Any]:
        return {}

    def to_response(self) -> Tuple[Any, int]:
        """
        Transforms this exception into a HTTP response containing a
        machine-readable description of the exception.
        """
        jsn = {
            'kind': self.__class__.__name__,
            'message': self.message
        }
        data = self.data
        if data:
            jsn['data'] = data
        jsn = {'error': jsn}
        return jsn, self.__status_code


class BadFormat(ClientServerError):
    """
    A given data structure did not match the expected format.
    """
    @staticmethod
    def from_data(data: dict) -> 'BadFormat':
        assert 'reason' in data
        return BadFormat(data['reason'])

    def __init__(self, reason: str) -> None:
        self.__reason = reason
        super().__init__(400, "unexpected format: {}".format(reason))

    @property
    def reason(self) -> str:
        return self.__reason

    @property
    def data(self) -> Dict[str, Any]:
        return {'reason': self.reason}


class UnexpectedServerError(ClientServerError):
    """
    The server encountered an unexpected error.
    """
    @staticmethod
    def from_exception(ex: Exception) -> 'UnexpectedServerError':
        return UnexpectedServerError(str(ex))

    @staticmethod
    def from_data(data: dict) -> 'UnexpectedServerError':
        assert 'message' in data
        return UnexpectedServerError(data['message'])

    def __init__(self, message: str) -> None:
        self.__message = message
        super().__init__(500, "unexpected error: {}".format(message))

    @property
    def message(self) -> str:
        return self.__message

    @property
    def data(self) -> Dict[str, Any]:
        return {'message': self.message}


class OperatorNameAlreadyExists(ClientServerError):
    """
    Used to indicate that a given operator name is already in use by another
    operator.
    """
    @staticmethod
    def from_data(data: dict) -> 'OperatorNameAlreadyExists':
        assert 'name' in data
        return OperatorNameAlreadyExists(data['name'])

    def __init__(self,
                 name: str,
                 *,
                 status_code: int = 409
                 ) -> None:
        self.__name = name
        msg = "operator name is already in use: {}".format(name)
        super().__init__(status_code, msg)

    @property
    def name(self) -> str:
        """
        The name of the requested operator.
        """
        return self.__name

    @property
    def data(self) -> Dict[str, Any]:
        return {'name': self.name}


class LanguageNotFound(ClientServerError):
    """
    Used to indicate that there exists no language registered under a given
    name.
    """
    @staticmethod
    def from_data(data: dict) -> 'LanguageNotFound':
        assert 'name' in data
        return LanguageNotFound(data['name'])

    def __init__(self,
                 name: str,
                 *,
                 status_code: int = 404
                 ) -> None:
        self.__name = name
        msg = "no language registered with name: {}".format(name)
        super().__init__(status_code, msg)

    @property
    def name(self) -> str:
        """
        The name of the requested language.
        """
        return self.__name

    @property
    def data(self) -> Dict[str, Any]:
        return {'name': self.name}


class LanguageNotDetected(ClientServerError):
    """
    Failed to automatically detect the language used by a file.
    """
    @staticmethod
    def from_data(data: dict) -> 'LanguageNotDetected':
        assert 'filename' in data
        return LanguageNotDetected(data['filename'])

    def __init__(self,
                 filename: str,
                 *,
                 status_code: int = 400
                 ) -> None:
        self.__filename = filename
        msg = "unable to automatically detect language used by file: {}"
        msg = msg.format(filename)
        super().__init__(status_code, msg)

    @property
    def filename(self) -> str:
        """
        The name of the file.
        """
        return self.__filename

    @property
    def data(self) -> Dict[str, Any]:
        return {'filename': self.filename}


class OperatorNotFound(ClientServerError):
    """
    Used to indicate that no operator has been registered under a given name.
    """
    @staticmethod
    def from_data(data: dict) -> 'OperatorNotFound':
        assert 'name' in data
        return OperatorNotFound(data['name'])

    def __init__(self,
                 name: str,
                 *,
                 status_code: int = 404
                 ) -> None:
        self.__name = name
        msg = "no operator registered with name: {}".format(name)
        super().__init__(status_code, msg)

    @property
    def name(self) -> str:
        """
        The name of the requested language.
        """
        return self.__name

    @property
    def data(self) -> Dict[str, Any]:
        return {'name': self.name}


class SnapshotNotFound(ClientServerError):
    """
    Used to indicate that no snapshot was found with the given name on the
    attached BugZoo server.
    """
    @staticmethod
    def from_data(data: dict) -> 'SnapshotNotFound':
        assert 'name' in data
        return SnapshotNotFound(data['name'])

    def __init__(self,
                 name: str,
                 *,
                 status_code: int = 404
                 ) -> None:
        self.__name = name
        msg = "no BugZoo snapshot registered with name: {}".format(name)
        super().__init__(status_code, msg)

    @property
    def name(self) -> str:
        """
        The name of the requested snapshot.
        """
        return self.__name

    @property
    def data(self) -> Dict[str, Any]:
        return {'name': self.name}


class FileNotFound(ClientServerError):
    """
    Used to indicate that the requested file was not found.
    """
    @staticmethod
    def from_data(data: dict) -> 'FileNotFound':
        assert 'name' in data
        return FileNotFound(data['name'])

    def __init__(self,
                 name: str,
                 *,
                 status_code: int = 404
                 ) -> None:
        self.__name = name
        msg = "file not found: {}".format(name)
        super().__init__(status_code, msg)

    @property
    def name(self) -> str:
        """
        The name of the missing file.
        """
        return self.__name

    @property
    def data(self) -> Dict[str, Any]:
        return {'name': self.name}


class BadConfigFile(BoggartException):
    """
    Used to indicate that a given configuration file is ill-formed.
    """
    def __init__(self, reason: str) -> None:
        super().__init__(reason)


class IllegalConfig(BoggartException):
    """
    Used to indicate that a given configuration is syntatically correct but
    that it describes an illegal configuration.
    """
    def __init__(self, reason: str) -> None:
        super().__init__(reason)



from typing import Dict, Any, List, FrozenSet, Iterator

from .transformation import Transformation
from .language import Language

__all__ = ['Operator']


class Operator(object):
    """
    Used to describe a mutation operator. Each mutation operator has its own
    own unique name that is used by external interfaces (i.e., the server) to
    lookup particular operators.
    """
    @staticmethod
    def from_dict(d: dict) -> 'Operator':
        assert 'name' in d
        assert 'languages' in d
        assert 'transformations' in d
        assert isinstance(d['name'], str)
        assert isinstance(d['languages'], list)
        assert isinstance(d['transformations'], list)
        assert all(isinstance(e, str) for e in d['languages'])
        assert all(isinstance(e, dict) for e in d['transformations'])

        name = d['name']
        languages = d['languages']
        transformations = \
            [Transformation.from_dict(t) for t in d['transformations']]

        return Operator(name, languages, transformations)

    def __init__(self,
                 name: str,
                 languages: List[str],
                 transformations: List[Transformation]
                 ) -> None:
        assert name != '', \
            "operator must have a non-empty name"
        assert languages != [], \
            "operator must support at least one language"
        assert transformations != [], \
            "operators must implement at least one transformation."

        self.__name = name
        self.__languages = frozenset(languages)
        self.__transformations = transformations.copy()

    @property
    def name(self) -> str:
        """
        The name of this mutation operator.
        """
        return self.__name

    @property
    def languages(self) -> Iterator[str]:
        """
        Returns an iterator over the names of the languages that are supported
        by this mutation operator.
        """
        return self.__languages.__iter__()

    @property
    def transformations(self) -> List[Transformation]:
        """
        The transformations performed by this mutation operator.
        """
        return self.__transformations.copy()

    def supports_language(self, language: Language) -> bool:
        """
        Determines whether this operator supports a given language.

        Returns:
            True if this operator supports the given language, else False.
        """
        return language.name in self.__languages

    def to_dict(self) -> Dict[str, Any]:
        """
        Provides a dictionary-based description of this transformation, ready
        to be serialized.
        """
        return {
            'name': self.name,
            'languages': list(self.languages),
            'transformations': [t.to_dict() for t in self.transformations]
        }



import logging
import boggart.exceptions
import boggart.warnings
import boggart.server
import boggart.client
import boggart.core

from .version import __version__
from .core import *
from .client import Client

logging.getLogger(__name__).addHandler(logging.NullHandler())
logging.getLogger(__name__).setLevel(logging.INFO)



from typing import List, Iterator, Dict, Any
from uuid import UUID

from .mutation import Mutation

__all__ = ['Mutant']


class Mutant(object):
    @staticmethod
    def from_dict(jsn: Any) -> 'Mutant':
        """
        Constructs a mutant description from a dictionary.
        """
        assert isinstance(jsn, dict)

        uuid = UUID(hex=jsn['uuid'])
        base = jsn['base']
        mutations = [Mutation.from_dict(m) for m in jsn['mutations']]

        return Mutant(uuid, base, mutations)

    def __init__(self,
                 uuid: UUID,
                 base: str,
                 mutations: List[Mutation]
                 ) -> None:
        """
        Constructs a new Mutant description.

        Parameters:
            uuid: the UUID for the mutant.
            base: the name of the snapshot that was used to generate the
                mutant.
            mutations: the sequence of mutations that were applied to the
                snapshot.
        """
        self.__uuid = uuid
        self.__base = base
        self.__mutations = mutations

    def __repr__(self) -> str:
        return "Mutant({}, {}, {})".format(self.uuid,
                                           self.base,
                                           repr(self.mutations))

    @property
    def uuid(self) -> UUID:
        """
        The UUID for this mutant.
        """
        return self.__uuid

    @property
    def base(self) -> str:
        """
        The name of the snapshot that was used to generate this mutant.
        """
        return self.__base

    @property
    def snapshot(self) -> str:
        """
        The name of the BugZoo snapshot for this mutant.
        """
        return "boggart:{}".format(self.uuid)

    @property
    def docker_image(self) -> str:
        """
        The name of the Docker image for this mutant.
        """
        return "boggart/{}".format(self.uuid)

    @property
    def mutations(self) -> Iterator[Mutation]:
        """
        Returns an iterator over the mutations that were applied to the
        snapshot in order to generate this mutant.
        """
        return self.__mutations.__iter__()

    def to_dict(self) -> Dict[str, Any]:
        """
        Produces a dictionary-based description of this mutant, ready to be
        serialized as, for instance, YAML or JSON.
        """
        return {
            'uuid': self.__uuid.hex,
            'base': self.__base,
            'mutations': [m.to_dict() for m in self.mutations]
        }



from typing import Any

import attr

from bugzoo.core.fileline import FileLine

__all__ = ['Location', 'LocationRange', 'FileLocationRange', 'FileLine']


@attr.s(frozen=True)
class Location(object):
    @staticmethod
    def from_string(s: str) -> 'Location':
        line, _, col = s.partition(':')
        return Location(int(line), int(col))

    line = attr.ib(type=int)
    col = attr.ib(type=int)

    @property
    def column(self) -> int:
        return self.col

    def __str__(self) -> str:
        return "{}:{}".format(self.line, self.column)


@attr.s(frozen=True)
class LocationRange(object):
    """
    Captures a continuous range of source code locations.
    """
    @staticmethod
    def from_string(s: str) -> 'LocationRange':
        start_s, _, stop_s = s.partition("::")
        start = Location.from_string(start_s)
        stop = Location.from_string(stop_s)
        return LocationRange(start, stop)

    start = attr.ib(type=Location)
    stop = attr.ib(type=Location)

    def __str__(self) -> str:
        return "{}::{}".format(self.start, self.stop)


@attr.s(frozen=True)
class FileLocationRange(object):
    """
    Represents a contiguous sequence of characters in a particular file.
    """
    @staticmethod
    def from_string(s: str) -> 'FileLocationRange':
        filename, _, s_range = s.rpartition('@')
        start_s, _, stop_s = s_range.partition('::')
        start = Location.from_string(start_s)
        stop = Location.from_string(stop_s)
        return FileLocationRange(filename, start, stop)

    filename = attr.ib(type=str)
    start = attr.ib(type=Location)
    stop = attr.ib(type=Location)

    def __str__(self) -> str:
        return "{}@{}::{}".format(self.filename, self.start, self.stop)



from typing import Dict, Union, List, Any
from urllib.parse import urljoin, urlparse
from timeit import default_timer as timer
import time
import logging

import requests

from ..exceptions import ConnectionFailure

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

__all__ = ['API']


class API(object):
    def __init__(self,
                 base_url: str,
                 *,
                 timeout: int = 30,
                 timeout_connection: int = 60
                 ) -> None:
        """
        Constructs a new low-level API client for communicating with a boggart
        server at a given address.

        Parameters:
            base_url: the base URL of the boggart server.
            timeout: the default timeout for API calls (in seconds).
            timeout_connection: the maximum numbers of seconds to wait when
                attempting to connect to the server before raising a
                ConnectionFailure exception.

        Raises:
            ValueError: if the provided URL lacks a scheme (e.g., 'http').
            ConnectionFailure: if a connection to the server could not be
                established within the connection timeout window.
        """
        if not urlparse(base_url).scheme:
            logger.error("invalid base URL provided: missing scheme.")
            raise ValueError("invalid base URL provided: missing scheme (e.g., 'http').")  # noqa: pycodestyle

        self.__base_url = base_url
        self.__timeout = timeout

        logger.info("attempting to establish connection to server '%s' with timeout of %d seconds",  # noqa: pycodestyle
                    base_url,
                    timeout)
        url = self.url("status")
        time_started = timer()
        connected = False
        while not connected:
            time_running = timer() - time_started
            time_left = timeout_connection - time_running
            if time_left <= 0.0:
                logger.error("failed to connect to server: %s", base_url)
                raise ConnectionFailure
            try:
                r = requests.get(url, timeout=time_left)
                connected = r.status_code == 204
                logger.info("connected to server: %s", base_url)
            except requests.exceptions.ConnectionError:
                time.sleep(1.0)
            except requests.exceptions.Timeout:
                logger.error("failed to connect to server: %s", base_url)
                raise ConnectionFailure
            time.sleep(0.05)

    @property
    def base_url(self) -> str:
        """
        The base URL of the server.
        """
        return self.__base_url

    def url(self, path: str) -> str:
        """
        Computes the URL to a resource located at a given path on the server.
        """
        return urljoin(self.__base_url, path)

    def get(self,
            path: str,
            params: Dict[str, Any] = None,
            **kwargs
            ) -> requests.Response:
        url = self.url(path)
        return requests.get(url, params, **kwargs)

    def post(self, path: str, data=None, **kwargs) -> requests.Response:
        url = self.url(path)
        return requests.post(url, data, **kwargs)

    def put(self, path: str, **kwargs) -> requests.Response:
        url = self.url(path)
        return requests.put(url, **kwargs)



from typing import Optional, List, Dict, Any, Iterator
from functools import wraps
from contextlib import contextmanager
from uuid import UUID
import argparse
import os
import signal
import subprocess
import logging
import logging.handlers
import sys

import bugzoo
import bugzoo.server
import flask
import rooibos
from flask_api import FlaskAPI

from .installation import Installation
from ..exceptions import *
from ..core import Language, Operator, Mutation
from ..client import Client

app = FlaskAPI(__name__)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# TODO: tidy this up
installation = None  # type: Any


@contextmanager
def ephemeral(*,
              url_bugzoo: str = 'http://127.0.0.1:6060',
              url_rooibos: str = 'http://127.0.0.1:8888',
              port: int = 8000,
              verbose: bool = False
              ) -> Iterator[Client]:
    """
    Launches an ephemeral server instance that will be immediately
    close when no longer in context.

    Parameters:
        port: the port that the server should run on.
        verbose: if set to True, the server will print its output to the
            stdout, otherwise it will remain silent.

    Returns:
        a client for communicating with the server.
    """
    url = "http://127.0.0.1:{}".format(port)
    cmd = ["boggartd",
           "-p", str(port),
           "--bugzoo", url_bugzoo,
           "--rooibos", url_rooibos]
    try:
        stdout = None if verbose else subprocess.DEVNULL
        stderr = None if verbose else subprocess.DEVNULL
        proc = subprocess.Popen(cmd,
                                preexec_fn=os.setsid,
                                stdout=stdout,
                                stderr=stderr)
        yield Client(url)
    finally:
        os.killpg(proc.pid, signal.SIGTERM)


@contextmanager
def ephemeral_stack(*,
                    port_boggart: int = 8000,
                    port_rooibos: int = 8888,
                    port_bugzoo: int = 6060,
                    verbose: bool = False
                    ) -> Iterator[Client]:
    """
    Launches an ephemeral server instance along with a complete underlying
    stack (i.e., an ephemeral Rooibos and BugZoo), all of which will be
    immediately destroyed upon leaving the context.

    Parameters:
        port_boggart: the port that the boggart server should run on.
        port_rooibos: the port that the Rooibos server should run on.
        port_bugzoo: the port that the BugZoo server should run on.
        verbose: if set to True, the server will print its output to the
            stdout, otherwise it will remain silent.

    Returns:
        a client for communicating with the boggart server.
    """
    url_rooibos = "http://127.0.0.1:{}".format(port_rooibos)
    url_bugzoo = "http://127.0.0.1:{}".format(port_bugzoo)
    with bugzoo.server.ephemeral(port=port_bugzoo,
                                 verbose=verbose) as client_bz:
        with rooibos.ephemeral_server(port=port_rooibos,
                                      verbose=verbose) as client_rooibos:
            with ephemeral(url_rooibos=url_rooibos,
                           url_bugzoo=url_bugzoo,
                           port=port_boggart,
                           verbose=verbose) as client_boggart:
                yield client_boggart


def throws_errors(func):
    """
    Wraps a function responsible for implementing an API endpoint such that
    any server errors that are thrown are automatically transformed into
    appropriate HTTP responses.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ClientServerError as err:
            logger.exception("encountered an error while handling request: %s", err.message)  # noqa: pycodestyle
            return err.to_response()
        except Exception as err:
            logger.exception("encountered unexpected error while handling request: %s", err)  # noqa: pycodestyle
            return UnexpectedServerError.from_exception(err).to_response()
    return wrapper


@app.route('/status', methods=['GET'])
@throws_errors
def status():
    """
    Produces a diagnostic summary of the health of the server.
    """
    logger.info("inspecting server health")
    return '', 204


@app.route('/languages/<name>', methods=['GET'])
@throws_errors
def describe_language(name: str):
    """
    Provides a description of a given language, specified by its name.

    Raises:
        LanguageNotFound: if no language is registered with this server
            under the given name.
    """
    logger.info("attempting to find language with name: %s", name)
    try:
        language = installation.languages[name]
        logger.info("found language with name %s: %s", name, language)
        return language.to_dict()
    except KeyError:
        logger.error("failed to find language: %s", name)
        raise LanguageNotFound(name)


@app.route('/languages', methods=['GET'])
def list_languages():
    """
    Produces a list of all languages that are supported by this server.
    """
    logger.info("retrieving list of supported languages")
    return [lang.to_dict() for lang in installation.languages]


@app.route('/operators/<name>', methods=['GET'])
@throws_errors
def describe_operator(name: str):
    """
    Describes a named operator.
    """
    logger.info("attempting to find operator with name: %s", name)
    try:
        operator = installation.operators[name]
        logger.info("found operator with name %s: %s", name, operator)
        return operator.to_dict()
    except KeyError:
        logger.error("failed to find operator: %s", name)
        raise OperatorNotFound(name)


@app.route('/operators', methods=['GET'])
@throws_errors
def list_operators():
    """
    Produces a list of all operators that are registered with boggart that
    satisfy a set of optionally provided parameters.

    Params:
        language: If supplied, restricts the set of operators to those that
            are compatible with a certain language, given by its name.
    """
    # TODO use URL params
    args = flask.request.get_json()
    if args is None:
        args = {}
    logger.info("requesting list of operators", extra=args)

    # get a list of all registered operators
    op_list = list(installation.operators)  # type: List[Operator]

    # perform optional language filtering
    if 'language' in args:
        try:
            language = installation.languages[args['language']]
        except KeyError:
            raise LanguageNotFound(args['language'])

        op_list = [op for op in op_list if op.supports_language(language)]

    # serialize to JSON
    jsn_op_list = [op.to_dict() for op in op_list]
    logger.info("operators that satisfy given constraints: [%s]",
                ', '.join([op.name for op in op_list]),
                extra={'operators': jsn_op_list})
    return jsn_op_list


@app.route('/diff/mutations/<name_snapshot>', methods=['PUT'])
@throws_errors
def mutations_to_diff(name_snapshot: str):
    """
    Transforms a set of mutations to a given snapshot into a unified diff.
    """
    sources = installation.sources
    logger.info("attempting to transform mutations into unified diff")
    logger.debug("attempting to retrieve snapshot: %s", name_snapshot)
    try:
        snapshot = installation.bugzoo.bugs[name_snapshot]
    except KeyError:
        logger.exception("failed to find snapshot: %s", name_snapshot)
        raise SnapshotNotFound(name_snapshot)

    logger.debug("extracting mutations from payload")
    try:
        mutations = \
            [Mutation.from_dict(m) for m in flask.request.json['mutations']]
    except KeyError:
        logger.exception("failed to transform mutations into unified diff: failed to read mutations from payload.")  # noqa: pycodestyle
        raise BadFormat("expected a JSON-encoded list of mutations")
    logger.debug("extracted mutations from payload")

    diff = sources.mutations_to_diff(snapshot, mutations)
    diff_s = str(diff)
    logger.info("transformed mutations into unified diff.")
    return diff_s, 200


@app.route('/mutants', methods=['GET', 'POST'])
@throws_errors
def interact_with_mutants():
    if flask.request.method == 'GET':
        logger.info("request list of mutants")
        list_uuid = [m.uuid for m in installation.mutants]
        logger.info("%d mutants: %s", len(list_uuid), list_uuid,
                    extra={'mutants': list_uuid})
        return flask.jsonify(list_uuid), 200

    if flask.request.method == 'POST':
        description = flask.request.json
        snapshot_name = description['snapshot']
        logger.info("generating mutant of snapshot '%s'",
                    snapshot_name,
                    extra={'payload': description})
        try:
            snapshot = installation.bugzoo.bugs[snapshot_name]
        except KeyError:
            logger.error("failed to generate mutant: snapshot (%s) was not found",  # noqa: pycodestyle
                         snapshot_name,
                         exc_info=True)
            return SnapshotNotFound(snapshot_name), 404

        mutations = [Mutation.from_dict(m) for m in description['mutations']]
        logger.debug("generating mutant of snapshot '%s' using mutations: %s",
                     snapshot_name, mutations)
        mutant = installation.mutants.generate(snapshot, mutations)
        logger.info("generated mutant: %s", mutant)
        jsn_mutant = mutant.to_dict()
        return flask.jsonify(jsn_mutant), 200


@app.route('/mutants/<uuid_str>', methods=['GET'])
@throws_errors
def interact_with_mutant(uuid_hex: str):
    uuid = UUID(hex=uuid_hex)

    if flask.request.method == 'GET':
        logger.info("fetching information about mutant: %s", uuid)
        try:
            mutant = installation.mutants[uuid]
        except Exception:
            logger.exception("failed to find mutant: %s", uuid)
        return flask.jsonify(mutant.to_dict()), 200


@app.route('/mutations/<name_snapshot>/<path:filepath>', methods=['GET'])
@throws_errors
def mutations(name_snapshot: str, filepath: str):
    """
    Determines the set of possible single-order mutations that can be applied
    to a given file belonging to a specified BugZoo snapshot.

    Path Parameters:
        name_snapshot: The name of the BugZoo snapshot that should be mutated.
        filepath: The file that should be analyzed.

    URL-encoded Parameters:
        language: An optional parameter that can be used to explicitly state
            the language used by the given file. If this parameter is not
            supplied, boggart will attempt to automatically detect the language
            used by a given file based on its file ending.
        operators: An optional semi-colon delimited parameter that can be used
            to specify which mutation operators should be used. If left
            unspecified, all available mutation operators for the language used
            by the given file will be used.

    Raises:
        SnapshotNotFound: if no snapshot can be found with the given name.
        FileNotFound: if the given file is not found inside the snapshot.
        LanguageNotFound: if the given language is not recognized by the
            server.
        LanguageNotDetected: if the language used by the given file cannot be
            automatically detected.
    """
    args = flask.request.args
    logger.info("attempting to find mutations to file '%s' belonging to snapshot '%s'",  # noqa: pycodestyle
                filepath,
                name_snapshot,
                extra={'arguments': args})

    # fetch the given snapshot
    logger.debug("attempting to retrieve snapshot: %s", name_snapshot)
    try:
        snapshot = installation.bugzoo.bugs[name_snapshot]
    except KeyError:
        logger.exception("failed to find snapshot: %s", name_snapshot)
        raise SnapshotNotFound(name_snapshot)
    logger.debug("retrieved snapshot: %s", name_snapshot)

    # determine the language used by the file
    if 'language' in args:
        logger.info("obtaining language from arguments")
        try:
            language = installation.languages[args['language']]
        except KeyError:
            raise LanguageNotFound(args['language'])
    else:
        logger.info("no language specified -- attempt to autodetect.")
        language = None

    # determine the set of operators that should be used
    if 'operators' in args:
        operators = []
        operator_names = args['operators'].split(';')
        logger.info("finding operators specified by argument: %s",
                    args['operators'],
                    extra={'operators': operator_names})
        for name in operator_names:
            try:
                logger.debug("looking for operator: %s", name)
                op = installation.operators[name]
                logger.debug("found operator with name %s: %s", name, op)
                operators.append(op)
            except KeyError:
                logger.exception("failed to find operator: %s", name)
                raise OperatorNotFound(name)
    else:
        logger.info("using all available operators to generate mutations")
        operators = list(installation.operators)

    # TODO implement line restriction
    try:
        generator_mutations = \
            installation.mutations(snapshot,
                                   filepath,
                                   language=language,
                                   operators=operators)
        mutations = list(generator_mutations)
    except BoggartException as e:
        logger.exception("failed to find mutations due to error: %s", e.message)  # noqa: pycodestyle
        raise
    except Exception as e:
        logger.exception("failed to find mutations due to unexpected error: %s", e)  # noqa: pycodestyle
        raise

    logger.info("found %d mutations of file '%s' in snapshot '%s' that satisfy the given constraints.",  # noqa: pycodestyle
                len(mutations),
                filepath,
                name_snapshot)

    logger.debug("serialising discovered mutations")
    jsn = [m.to_dict() for m in mutations]
    logger.debug("serialised discovered mutations")
    return jsn


def launch(port: int = 8000,
           url_bugzoo: str = 'http://127.0.0.1:6060',
           url_rooibos: str = 'http://host.docker.internal:8888',
           host: str = '0.0.0.0',
           debug: bool = False,
           log_filename: Optional[str] = None
           ) -> None:
    global installation

    log_formatter = \
        logging.Formatter('%(asctime)s:%(name)s:%(levelname)s: %(message)s',
                          '%Y-%m-%d %H:%M:%S')

    if not log_filename:
        log_filename = "boggartd.log"
        log_filename = os.path.join(os.getcwd(), log_filename)

    log_to_file = logging.handlers.WatchedFileHandler(log_filename, mode='w')
    log_to_file.setLevel(logging.DEBUG)
    log_to_file.setFormatter(log_formatter)

    log_to_stdout = logging.StreamHandler()
    log_to_stdout.setLevel(logging.INFO)
    log_to_stdout.setFormatter(log_formatter)

    log_main = logging.getLogger('boggart')  # type: logging.Logger
    log_main.addHandler(log_to_stdout)
    log_main.addHandler(log_to_file)

    assert 0 <= port <= 49151

    logger.info("attempting to connect to BugZoo server: %s", url_bugzoo)
    client_bugzoo = bugzoo.client.Client(url_bugzoo)
    logger.info("connected to BugZoo server")
    logger.info("attempting to connect to Rooibos server: %s",
                url_rooibos)
    client_rooibos = rooibos.Client(url_rooibos, timeout_connection=60)
    logger.info("connected to Rooibos server")
    installation = Installation.load(client_bugzoo, client_rooibos)
    logger.info("launching HTTP server at %s:%d", host, port)
    app.run(port=port, host=host, debug=debug)


def main() -> None:
    desc = 'boggart'
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument('-p', '--port',
                        type=int,
                        default=8000,
                        help='the port that should be used by this server.')
    parser.add_argument('--log-file',
                        type=str,
                        help='the path to the file where logs should be written.')  # noqa: pycodestyle
    parser.add_argument('--rooibos',
                        type=str,
                        default='http://host.docker.internal:8888',
                        help='the URL of the Rooibos server.')
    parser.add_argument('--bugzoo',
                        type=str,
                        default='http://127.0.0.1:6060',
                        help='the URL of the BugZoo server.')
    parser.add_argument('--host',
                        type=str,
                        default='0.0.0.0',
                        help='the IP address of the host.')
    parser.add_argument('--debug',
                        action='store_true',
                        help='enables debugging mode.')
    args = parser.parse_args()

    launch(port=args.port,
           url_bugzoo=args.bugzoo,
           url_rooibos=args.rooibos,
           host=args.host,
           debug=args.debug,
           log_filename=args.log_file)



from typing import Any, Dict

__all__ = ['Transformation']


class Transformation(object):
    """
    Describes a source code transformation as a corresponding pair of Rooibos
    match and rewrite templates.
    """
    @staticmethod
    def from_dict(d: dict) -> 'Transformation':
        assert 'match' in d
        assert 'rewrite' in d
        assert isinstance(d['match'], str)
        assert isinstance(d['rewrite'], str)

        match = d['match']
        rewrite = d['rewrite']

        return Transformation(match, rewrite)

    def __init__(self, match: str, rewrite: str) -> None:
        self.__match = match
        self.__rewrite = rewrite

    def __repr__(self) -> str:
        return "Transformation({}, {})".format(self.match, self.rewrite)

    def __hash__(self) -> int:
        return hash((self.match, self.rewrite))

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Transformation) and \
               self.match == other.match and \
               self.rewrite == other.rewrite

    @property
    def match(self) -> str:
        return self.__match

    @property
    def rewrite(self) -> str:
        return self.__rewrite

    def to_dict(self) -> dict:
        """
        Provides a dictionary--based description of this transformation, ready
        to be serialized.
        """
        return {'match': self.__match,
                'rewrite': self.__rewrite}
